from jieba import posseg
import jieba
from collections import Counter
import pandas as pd
from word_expanding.utilities import get_file_lines
from tqdm import tqdm


class TextManager:

    def __init__(self, dict=None):
        self.__cutter = jieba
        self.__posseg = posseg
        if dict is not None: self.__cutter.load_userdict(dict)
        self.all_words = Counter()

    def get_stopwords(self, stop_words_size=20):
        return self.all_words.most_common(n=stop_words_size)

    def cut(self, string):
        words = [w.upper() for w in list(self.__cutter.cut(string))]
        self.all_words.update(words)
        return words

    def get_split_line(self, string): return ' '.join(self.cut(str(string).strip()))

    @staticmethod
    def write_to_file(line, f):
        f.write('{}\n'.format(line))

    def append_text_file(self, new_file, old_file):
        try:
            file_lines = get_file_lines(new_file)
        except Exception as e:
            file_lines = None

        with open(new_file, 'r', encoding='utf-8') as f:
            if file_lines is not None:
                file_writer = tqdm(enumerate(f), total=file_lines)
            else:
                file_writer = tqdm(enumerate(f))

            for ii, line in file_writer:
                self.write_to_file(self.get_split_line(line), old_file)
                file_writer.set_description(desc=str(new_file))
                file_writer.set_postfix(line=ii)

    def append_csv_file(self, csv_file, old_file):
        content = pd.read_csv(csv_file, encoding='utf-8')
        rows = len(content)
        columns_n = len(content.columns)

        tqdm_iterator = tqdm(enumerate(content.iterrows()), total=int(rows))
        for ii, row in tqdm_iterator:
            for column in self.get_valid_columns(row[1]):
                self.write_to_file(self.get_split_line(column), old_file)
            tqdm_iterator.update(columns_n)
            tqdm_iterator.set_description(desc=str(csv_file))
            tqdm_iterator.set_postfix(line=ii)

    @staticmethod
    def get_valid_columns(row):
        return filter(TextManager.is_valid_content, row.tolist())

    @staticmethod
    def is_valid_content(s):
        conds = [
            TextManager.is_pure_person_name,
            TextManager.is_pure_number
        ]
        return not any([cond(str(s)) for cond in conds])

    @staticmethod
    def is_pure_number(s): return str(s).isnumeric()

    @staticmethod
    def is_person_name(t): return t.startswith('nr')

    @staticmethod
    def is_pure_person_name(string):
        return all([TextManager.is_person_name(t) for w, t in posseg.cut(string) if t != 'x'])


if __name__ == '__main__':
    assert TextManager.is_pure_person_name('李英杰')
    assert not TextManager.is_pure_person_name('李英杰是个好人')
    assert TextManager.is_pure_person_name('"李英杰","段涵洲"')
    assert TextManager.is_pure_number('1321421312890')
    assert TextManager.is_valid_content('2088202933720187') is False
    assert TextManager.is_valid_content('万代高达模型工具通用展示台支架底座mg rg hg敢达配件1/144 100') is True
    assert TextManager.is_valid_content('李英杰') is False

    W = TextManager(dict='data/dict.txt')

    s = '万代高达模型工具通用展示台支架底座mg'
    assert W.cut(s) == ['万代', '高达', '模型', '工具', '通用', '展示台', '支架', '底座', 'MG'], W.cut(s)



import hashlib
import os
import subprocess


def md5(fname):
    hash_md5 = hashlib.md5()
    with open(fname, 'rb') as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def connect_words(words, connecter='-'):
    return connecter.join(map(str, words))


def hash_string(string):
    return hashlib.md5(string.encode()).hexdigest()


def get_file_lines(file):
    if os.name == 'nt':  # window operation system cannot call wc -l
        res = 0
        with open(file):
            for _ in file: res +=1
        return res
    else:
        p = subprocess.Popen(['wc', '-l', file], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        result, err = p.communicate()
        if p.returncode != 0:
            raise IOError(err)
        return int(result.strip().split()[0])


def logging_info(key_words):
    print('##' * 22 + ' ' + key_words.upper() + ' ' + '##' * 22)


if __name__ == '__main__':
    assert md5('data/dict.txt') == md5('../data3.txt')
    assert md5('data/dict.txt') != md5('../data_2.txt')

    assert hash_string('test1') == hash_string('test1')
    assert hash_string('test1') != hash_string('test2')

    assert get_file_lines('data/dict.txt') == 1, get_file_lines('data/dict.txt')
    logging_info('embedding')



from gensim.models import KeyedVectors
from datetime import datetime
from collections import defaultdict
from tqdm import tqdm
import csv
from word_expanding.text_manager import TextManager


def expanding_words(initial_words_file, expanding_time, word2vec_model, mark="", stopwords=None):
    now = datetime.now().strftime('%Y-%m-%d-%H:%M:%S:%s')[:-8]
    model = KeyedVectors.load_word2vec_format(word2vec_model)

    initial_words = [line.strip() for line in open(initial_words_file, encoding='utf-8')]
    already_know = [w.upper() for w in initial_words]
    unprocessed = [w for w in already_know if w in model.vocab]
    occurence = defaultdict(int)

    calculated = set()

    tqdm_iterator = tqdm(range(expanding_time), total=expanding_time)

    while len(unprocessed) > 0 and expanding_time >= 0:
        w = unprocessed.pop()
        tqdm_iterator.set_description('Iter: {}'.format(expanding_time))
        if w in calculated: continue
        most_similar = [w for w, v in model.wv.most_similar(w)]
        for _w in most_similar:
            tqdm_iterator.set_postfix(new_word=_w)
            occurence[_w] += 1
        unprocessed += most_similar
        expanding_time -= 1
        tqdm_iterator.update(1)
        calculated.add(w)

    occurence = sorted(occurence.items(), key=lambda x: x[1], reverse=True)

    def contain_in_initial(w):
        for iw in already_know:
            if iw.find(w) >= 0 or w.find(iw) >= 0: return True
        return False

    suspected = [(w, c) for w, c in occurence[:100] if not contain_in_initial(w) and w not in stopwords]
    suspected = [(w, c) for w, c in suspected if TextManager.is_valid_content(w)]

    with open('{}-{}-expanding.csv'.format(now, mark), 'w', encoding='utf-8') as f:
        writer = csv.writer(f)
        writer.writerow(['word', 'value'])
        for w, c in suspected:
            writer.writerow([w, c])

    print('expanding end!')






"""
Training Word Embedding
"""

import jieba
import os
from word_expanding import utilities
from word_expanding.text_manager import TextManager
import fasttext
from word_expanding.utilities import connect_words
from word_expanding.utilities import hash_string
from word_expanding.utilities import logging_info
from word_expanding.expanding import expanding_words

SRC_DIR = './source_files'
if not os.path.exists(SRC_DIR):
    print('create source_file dir')
    os.mkdir(SRC_DIR)

TGT_DIR = './targets'

if not os.path.exists(TGT_DIR): os.mkdir(TGT_DIR)


class Embedding:
    def __init__(self, input_files=None, initial_words_file=None, dim=50, expanding_time=50, method='skip-gram', self_define_dictionary=None, update_corpus=False):
        """
        :param dim: the dimonsion of word2vec, default is 50.
        :param expanding_time: the expanding times, default is 50, the more expanding time, the further model search.
        :param method: skip-gram or cbow, default is skip-gram
        :param self_define_dictionary: self-defined dictionay file name, e.g 'your-self-dict.txt'
        :param update_corpus: Boolean, if is true, will update all the train corpus.

        e.g
            files = [
                root + 'gamble_paid.csv',
                root + 'group_names.txt',
            ]

            embedding = Embedding(input_files=files,
                                initial_words_file='../initial_words.txt',
                                dim=50, expanding_time=100, update_corpus=False)
            embedding.expanding_new_words()
        """
        if self_define_dictionary is not None: jieba.load_userdict(self_define_dictionary)

        initial_words_file = initial_words_file or './initial_words.txt'

        if not os.path.exists(initial_words_file):
            raise NameError('please put your initial words in initial_words.txt')

        src_dir = './source_files'
        input_files = input_files or ["{}/{}".format(src_dir, f) for f in os.listdir(src_dir)]

        if len(input_files) == 0: raise TypeError('source files is None, please put your files into ./source_files')

        target_dir = './targets'
        if not os.path.exists(target_dir):
            os.mkdir(target_dir)

        self.target_dir = target_dir
        self.dict = self_define_dictionary
        self.initial_words_file = initial_words_file
        self.method = method
        self.input_files = input_files
        self.dim = dim
        self.text_manager = TextManager(self_define_dictionary)
        self.update_corpus = update_corpus
        self.expanding_time = expanding_time

    def expanding_new_words(self):
        if self.embedding_exist():
            self.expanding()
        else:
            self.get_embedding()
            self.expanding_new_words()

    def expanding(self):
        expanding_words(
            self.initial_words_file,
            self.expanding_time,
            self.get_w2v_model_name()+'.vec',
            mark=str(self.dim) + '-' + str(self.expanding_time),
            stopwords=self.text_manager.get_stopwords(),
        )

    def get_embedding(self):
        if self.update_corpus is True or not self.line_corpus_exist():
            self.get_line_corpus()

        assert os.path.exists(self.get_line_corpus_name())

        if self.method.lower() == 'skip-gram':
            embedding = fasttext.skipgram
        elif self.method.lower() == 'cbow':
            embedding = fasttext.cbow
        else:
            raise TypeError('Unsupported embedding method, <skip-gram(default), cbow> are supported!')

        logging_info('training embedding')
        embedding(self.get_line_corpus_name(), self.get_w2v_model_name(), dim=self.dim)
        logging_info('end embedding')

    def get_line_corpus_name(self):
        line_corpus_file = hash_string(connect_words([self.input_files, self.dict])) + '.txt'
        return self.target_dir + '/' + line_corpus_file

    def get_w2v_model_name(self):
        if self.dict is not None and os.path.exists(self.dict):
            dict_md5 = utilities.md5(self.dict)
        else:
            dict_md5 = ""

        w2v_model_name = hash_string(connect_words([self.input_files, self.dim, self.method, dict_md5]))

        return self.target_dir + '/' + w2v_model_name

    def line_corpus_exist(self):
        return os.path.exists(self.get_line_corpus_name())

    def embedding_exist(self):
        return os.path.exists(self.get_w2v_model_name() + '.vec')

    def get_line_corpus(self):
        logging_info('get training corpus')
        with open(self.get_line_corpus_name(), 'w', encoding='utf-8') as fline:
            self.input_files = [f.strip() for f in self.input_files]
            for file in self.input_files:
                if file.endswith('.csv'):
                    self.text_manager.append_csv_file(file, fline)
                else: self.text_manager.append_text_file(file, fline)


if __name__ == '__main__':
    embedding = Embedding(dim=80, expanding_time=60)
    embedding.expanding_new_words()



from distutils.core import setup

classifiers = [
    'Programming Language :: Python :: 3',
]

requirements = [
    'tqdm',
    'fasttext',
    'pandas',
    'numpy',
    'jieba',
]

paramters = {
    'name': 'word_expanding',
    'packages': ['word_expanding'],
    'version': '0.02',
    'author': 'minchiuan.gao',
    'author_email': 'minchiuan.gao@gmail.com',
    'url': 'https://github.com/fortymiles',
    'description': 'test for packaging',
    'install_requires': requirements
}


setup(**paramters)
